package be.iminds.ilabt.jfed.experiment;

import be.iminds.ilabt.jfed.call_log_output.LogOutput;
import be.iminds.ilabt.jfed.espec.model.ESpecStep;
import be.iminds.ilabt.jfed.experiment.events.*;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.experiment.util.NextExperimentExpiration;
import be.iminds.ilabt.jfed.experiment.util.NextExperimentExpirationBinding;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Service;
import be.iminds.ilabt.jfed.highlevel.jobs.*;
import be.iminds.ilabt.jfed.highlevel.jobs.report.JobReport;
import be.iminds.ilabt.jfed.highlevel.jobs.report.StitchingJobReport;
import be.iminds.ilabt.jfed.highlevel.model.InternalState;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.highlevel.util.LogEntryGeneratorWrappingLogger;
import be.iminds.ilabt.jfed.highlevel.util.LogEntryListener;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.testbed_info.TestbedInfoSource;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.preferences.JFedCorePreferences;
import be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl.FXRspecLink;
import be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import be.iminds.ilabt.jfed.util.common.ThreadFactoryUtil;
import javafx.application.Platform;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.time.Duration;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.*;
import java.util.stream.Collectors;

/**
 * This class will take care of an Experiment-object. It makes sure that the appropriate actions are taken when
 * the state of the Slice, or one of it's slivers changes.
 * <p/>
 * User: twalcari
 * Date: 1/30/2015
 * Time: 14:04
 */
public class ExperimentControllerImpl implements ExperimentChangeListener, ExperimentController, LogEntryListener {
    private static final Logger ORIG_LOG = LoggerFactory.getLogger(ExperimentController.class);
    private final LogEntryGeneratorWrappingLogger LOG =
            new LogEntryGeneratorWrappingLogger(ORIG_LOG, this);


    @Nonnull private final Experiment experiment;
    private final ExperimentEventHandlerManager eventHandlerManager = new ExperimentEventHandlerManager();

    private final JobFactory jobFactory;
    private final TestbedInfoSource testbedInfoSource;
    private final JFedCorePreferences jFedPreferences;
    private final GeniUserProvider geniUserProvider;

    private final ScheduledExecutorService scheduledExecutorService = Executors.newScheduledThreadPool(2,
            ThreadFactoryUtil.getFactory("ExperimentController-Scheduled"));


    private static final ExecutorService jobExecutorService = Executors.newCachedThreadPool(
            ThreadFactoryUtil.getFactory("JobExecutor"));

    //helper variable for showing dialog messages. Ideally, this should be refactored away

    private final List<ExperimentControllerListener> listeners = new ArrayList<>();
    private final ExpirationDetector expirationDetector;

    private ExperimentConnectivityTesterFactory experimentConnectivityTesterFactory = null;
    private ExperimentLinkTesterFactory experimentLinkTesterFactory = null;
    private WaitForReadyTimeoutHandler waitForReadyTimeoutHandler = new DefaultWaitForReadyTimeoutHandler();

    public ExperimentControllerImpl(@Nonnull Experiment experiment,
                                    @Nonnull JobFactory jobFactory,
                                    @Nonnull TestbedInfoSource testbedInfoSource,
                                    @Nonnull JFedCorePreferences jFedPreferences,
                                    @Nonnull GeniUserProvider geniUserProvider) {
        this.experiment = experiment;
        this.jobFactory = jobFactory;
        this.testbedInfoSource = testbedInfoSource;
        this.jFedPreferences = jFedPreferences;
        this.geniUserProvider = geniUserProvider;

        this.expirationDetector = new ExpirationDetector();

        experiment.addExperimentChangeListener(this);
    }

    /**
     * This method must be called for the experimentController to start taking actions.
     */
    @Override
    public void start() {
        //bootstrap state machine
        LOG.debug("Bootstrapping experiment controller of '{}' in state {}", experiment.getName(), experiment.getExperimentState());
        onExperimentStateChange(experiment.getExperimentState());
    }


    @Override
    public void onExperimentStateChange(@Nonnull ExperimentState newExperimentState) {
        LOG.debug("onExperimentStateChange({})", newExperimentState);

        switch (newExperimentState) {
            case PENDING:
                createSlice();
                //next: CREATED
                break;
            case RESTORING:
                restoreSlice();
                //next: RESTORED or RENEW_EXISTING  (depending on if it's a recover or a new experiment creation)
                break;
            case RENEW_EXISTING:
                renewExistingSlice();
                //next: CREATED
                break;
            case RESTORED:
                handleRestored();
                //next: FUTURE_RESERVATION, WAIT_FOR_READY, TESTING_CONNECTIVITY or FAILED
                break;
            case CREATED:
                //startSliceExpirationDetection();
                assert experiment.getNewRequestRspecSource() != null;
                if (experiment.getRequestedStartTime() != null) {
                    if (!experiment.isReservationMade()) {
                        experiment.setExperimentState(ExperimentState.ALLOCATING);
                    } else if (experiment.getRequestedStartTime().isAfter(Instant.now())) {
                        experiment.setExperimentState(ExperimentState.FUTURE_RESERVATION);
                    } else {
                        // the reservation allocation was already done, and the start time has passed, so start provisioning
                        experiment.setExperimentState(ExperimentState.PROVISIONING);
                    }
                } else {
                    experiment.setExperimentState(ExperimentState.ALLOCATING);
                }
                break;
            case ALLOCATING:
                allocateResources();
                //next: FUTURE_RESERVATION, PROVISIONING or FAILING
                break;
            case FUTURE_RESERVATION:
                waitForReservationStart();
                //next: PROVISIONING  (after delay)
                break;
            case PROVISIONING:
                provisionResources();
                //next: WAIT_FOR_READY or FAILING
                break;
            case WAIT_FOR_READY:
                waitForReady();
                //next: TESTING_CONNECTIVITY or FAILING
                break;
            case TESTING_CONNECTIVITY:
                testSliversConnectivity(true);
                //next: TESTING_LINKS or FAILING
                break;
            case TESTING_LINKS:
                testLinkConnectivity();
                //next: SETUP_SOFTWARE (if create), READY (if recover) or FAILING
                break;
            case SETUP_SOFTWARE:
                setupSoftware();
                //next: READY or FAILING
                break;
            case READY:
                //final state (until user takes an action, or expiration)
                break;
            case TIMEOUT_WAITING:
                //user must make decision
                submitJob(jobFactory.createWaitForReadyTimeoutHandlerJob(experiment, waitForReadyTimeoutHandler));
                break;
            case FAILING:
                //failure, and currently asking user what to do
                //after user decision, goes to FAILING
                break;
            case FAILED:
                //final state
                break;
            case UNKNOWN:
                //final state
                break;
            case EXPIRING:
                checkForExpiration();
                break;
            case EXPIRED:
            case EMPTY:
                deleteRestoreInformation();
                break;
        }


    }

    private void deleteRestoreInformation() {
        experiment.getExperimentRestoreInformation().delete();
    }

    @Override
    public void onExperimentPartAdded(ExperimentPart experimentPart) {
        //do nothing
    }

    private void checkForExpiration() {
        requestUpdate();
    }

    private void checkExperimentPartsStates() {
        boolean allPartsDeletedOrExpired = experiment.getPartsListCopy().stream()
                .filter(Objects::nonNull)
                .filter(ep -> ep.getState() != null)
                .allMatch(ep -> ep.getState().equals(InternalState.DELETED) ||
                        ep.getState().equals(InternalState.UNALLOCATED));

        if (allPartsDeletedOrExpired) {
            if (experiment.getSliceOrNull() != null
                    && experiment.getSliceOrNull().getExpirationDate() != null
                    && experiment.getSliceOrNull().getExpirationDate().isBefore(Instant.now()))
                experiment.setExperimentState(ExperimentState.EXPIRED);
            else
                experiment.setExperimentState(ExperimentState.EMPTY);
        } else {
            boolean hasAnyFailingSliver = experiment.getSliceOrNull() != null &&
                    experiment.getSliceOrNull().getSliversStream()
                            .anyMatch(Sliver::hasAnyFailStatus);
            if (hasAnyFailingSliver) {
                LOG.debug("Setting experiment to FAILING due to failing sliver(s): {}",
                        experiment.getSliceOrNull().getSliversStream()
                                .filter(Sliver::hasAnyFailStatus)
                                .map(s -> s.getUrnString()+" with state "+s.getStatusString())
                                .collect(Collectors.toList()));
                experiment.setExperimentState(ExperimentState.FAILING);
            } else
                experiment.setExperimentState(ExperimentState.READY);
        }
    }

    public CompletableFuture<TestConnectivityJob.TestConnectivityJobResult> testConnectivity() {
        return testSliversConnectivity(false);
    }

    private CompletableFuture<TestConnectivityJob.TestConnectivityJobResult> testSliversConnectivity(boolean autoNextState) {
        if (experimentConnectivityTesterFactory != null) {
            return submitJob(experimentConnectivityTesterFactory.createTestExperimentConnectivityJob(experiment))
                    .whenCompleteAsync((success, throwable) -> {
                        NextExperimentExpiration nextExperimentExpiration = new NextExperimentExpiration(experiment);
                        if (nextExperimentExpiration.hasPartExpired()) {
                            experiment.setExperimentState(ExperimentState.EXPIRING);
                            return;
                        }

                        if (autoNextState) {
                            //always go to TESTING_LINKS. If it is skipped, it will go to next step automatically.
                            experiment.setExperimentState(ExperimentState.TESTING_LINKS);
                        }
                    }, Platform::runLater);

        } else {
            LOG.info("Skipping testing of experiment, as no ExperimentConnectivityTesterFactory is registered.");
            if (autoNextState) {
                experiment.setExperimentState(ExperimentState.TESTING_LINKS);
            }
            return CompletableFuture.completedFuture(TestConnectivityJob.TestConnectivityJobResult.createFalseEmpty());
        }
    }

    private static boolean hasLinks(Experiment experiment) {
        for (ExperimentPart part : experiment.getPartsListCopy()) {
            if (part instanceof SfaExperimentPart) {
                for (Sliver sliver : ((SfaExperimentPart) part).getSlivers()) {
                    if (sliver.getLinks() != null && !sliver.getLinks().isEmpty())
                        return true;
                }
            }
        }
        return false;
    }

    private void testLinkConnectivity() {
        if (!experiment.isLinkTestRequested()) {
            LOG.info("Skipping Link Test: Not requested.");
            deriveAndSetExperimentStateAfterTestingLinks();
        } else if (!hasLinks(experiment)) {
            LOG.info("Skipping Link Test: No links.");
            deriveAndSetExperimentStateAfterTestingLinks();
        } else if (experimentLinkTesterFactory == null) {
            LOG.info("Skipping link testing of experiment, as no ExperimentLinkTesterFactory is registered.");
            deriveAndSetExperimentStateAfterTestingLinks();
        } else {
            submitJob(experimentLinkTesterFactory.createTestLinkJob(experiment))
                    .whenCompleteAsync((aBoolean, throwable) -> deriveAndSetExperimentStateAfterTestingLinks(), Platform::runLater);
        }
    }

    private void deriveAndSetExperimentStateAfterTestingLinks() {
        assert experiment.getExperimentState() == ExperimentState.TESTING_LINKS :
                "experiment.getExperimentState() is not TESTING_LINKS but " + experiment.getExperimentState();

        NextExperimentExpiration nextExperimentExpiration = new NextExperimentExpiration(experiment);
        if (nextExperimentExpiration.hasPartExpired()) {
            experiment.setExperimentState(ExperimentState.EXPIRING);
            return;
        }

        boolean allReady = true;
        boolean allEmpty = true;

        for (ExperimentPart part : experiment.getPartsListCopy()) {
            LOG.debug("ExperimentController deriveAndSetExperimentStateAfterTesting part.getState()=" + part.getState());
            allReady = allReady && part.getState() == InternalState.READY;
            allEmpty = allEmpty && part.getState() == InternalState.UNALLOCATED;
        }

        LOG.debug("ExperimentController deriveAndSetExperimentStateAfterTesting allReady=" + allReady + " allEmpty=" + allEmpty);

        if (allReady) {
            if (experiment.getNewRequestRspecSource() != null) {
                experiment.setExperimentState(ExperimentState.SETUP_SOFTWARE);
            } else {
                experiment.setExperimentState(ExperimentState.READY);
            }
        } else {
            if (allEmpty) {
                experiment.setExperimentState(ExperimentState.EMPTY);
            } else {
                LOG.warn("Could not determine a valid experiment state for {}. Defaulting to allReady", experiment.getName());

                if (experiment.getNewRequestRspecSource() != null) {
                    experiment.setExperimentState(ExperimentState.SETUP_SOFTWARE);
                } else {
                    experiment.setExperimentState(ExperimentState.READY);
                }
            }
        }

        LOG.debug("ExperimentController deriveAndSetExperimentStateAfterTesting post state = " + experiment.getExperimentState());
    }


    @Override
    public CompletableFuture<Void> createSlice() {
        //we must initiate the creation of the slice for this experiment
        assert (experiment.getRequestedEndTime() != null);

        return submitJob(jobFactory.createRegisterExperimentJob(experiment));
    }

    private void restoreSlice() {
        submitJob(jobFactory.createRestoreExperimentJob(experiment));
    }

    private void renewExistingSlice() {
        if (experiment.getRequestedEndTime() == null) {
            experiment.setExperimentState(ExperimentState.CREATED);
            return;
        }

        NextExperimentExpiration nextExperimentExpiration = new NextExperimentExpiration(experiment);
        if (nextExperimentExpiration.getFirstExpirationTime() == null ||
                nextExperimentExpiration.getFirstExpirationTime().isBefore(experiment.getRequestedEndTime())) {

            //do renew
            submitJob(jobFactory.createRenewSliceJob(experiment, experiment.getRequestedEndTime()))
                    .whenCompleteAsync((renewStatus, throwable) -> {
                        if (throwable == null) {
                            if (renewStatus == ExperimentTaskStatus.SUCCESS) {
                                experiment.setExperimentState(ExperimentState.CREATED);
                            } else {
                                LOG.debug("Renew Slice for {} job success but renew failed: returned renewStatus={}",
                                        experiment.getName(), renewStatus);
                                experiment.setExperimentState(ExperimentState.FAILING);
                            }
                        } else {
                            LOG.error("Renew Slice for " + experiment.getName() + " failed", throwable);
                            experiment.setExperimentState(ExperimentState.FAILING);
                        }
                    }, Platform::runLater);
        } else {
            experiment.setExperimentState(ExperimentState.CREATED);
        }
    }

    private void handleRestored() {
        if (experiment.getRequestedStartTime() != null
                && experiment.getRequestedStartTime().isAfter(Instant.now())) {
            experiment.setExperimentState(ExperimentState.FUTURE_RESERVATION);
        } else {
            boolean hasFailure = experiment.getPartsStream()
                    .anyMatch(part -> part.getState() == InternalState.FAILED);

            boolean hasChanging = experiment.getPartsStream()
                    .filter(Objects::nonNull)
                    .map(ExperimentPart::getState)
                    .filter(Objects::nonNull)
                    .anyMatch(partState -> {
                        switch (partState) {
                            case CHANGING:
                            case ALLOCATING:
                            case PROVISIONING:
                            case DELETING:
                            case RESERVING:
                            case UNKNOWN:
                            case WAIT_FOR_READY:
                                return true;
                            default:
                                return false;
                        }
                    });

            if (hasFailure) {
                //not FAILING: we don't need to ask user
                experiment.setExperimentState(ExperimentState.FAILED);
                return;
            }

            if (hasChanging) {
                experiment.setExperimentState(ExperimentState.WAIT_FOR_READY);
                return;
            }

            experiment.setExperimentState(ExperimentState.TESTING_CONNECTIVITY);
        }
    }

    private void allocateResources() {
        if (experiment.getSliceOrNull() != null) {
            experiment.getSlice().setRequestRspec(experiment.getNewRequestRspecSource());
        }

        if (experiment.getExperimentSpecification() != null) {
            experiment.getExperimentSpecificationLogger().firePreRSpec();

            //noinspection ConstantConditions
            if (experiment.getExperimentSpecification().getRspecs() != null &&
                    !experiment.getExperimentSpecification().getRspecs().isEmpty()) {
                //RSpec is already known, it is fetched in Experiment constructor

                //eSpec normally uses newRequestRspecSource
                assert experiment.getNewRequestRspecSource() != null;
                assert experiment.getExistingRequestRspecSource() == null;

                if (experiment.getNewRequestRspecSource() != null &&
                        experiment.getNewRequestRspecSource().getRspecXmlString() != null) {
                    experiment.getExperimentSpecificationLogger().fireRequestRSpecKnown(
                            experiment.getExperimentSpecification().getRspecs().get(0),
                            experiment.getNewRequestRspecSource().getRspecXmlString()
                    );
                }
            }
        }

        AllocateExperimentJob allocateExperimentJob = jobFactory.createAllocateExperimentJob(experiment, getScsAuthority());
        StitchingJobReport stitchingJobReport = allocateExperimentJob.getStitchingJobReport();
        experiment.addJobReport(stitchingJobReport);
        submitJob(allocateExperimentJob)
                .whenCompleteAsync((success, throwable) -> {
                    if (throwable != null) {
                        LOG.error("AllocateExperimentJob for " + experiment.getName() + " failed: ", throwable);
                        experiment.setExperimentState(ExperimentState.FAILING);
                    } else {
                        if (Objects.equals(success, Boolean.TRUE)) {
                            if (experiment.getRequestedStartTime() == null || experiment.getRequestedStartTime().isBefore(Instant.now())) {
                                experiment.setExperimentState(ExperimentState.PROVISIONING);
                            } else {
                                experiment.setExperimentState(ExperimentState.FUTURE_RESERVATION);
                            }
                        } else {
                            LOG.debug("AllocateExperimentJob for {} succeeded but returned success={}", experiment.getName(), success);
                            experiment.setExperimentState(ExperimentState.FAILING);
                        }
                    }
                }, Platform::runLater);
    }


    private void waitForReservationStart() {
        assert (experiment.getRequestedStartTime() != null);

        Duration durationToStartTime = Duration.between(Instant.now(), experiment.getRequestedStartTime());

        Runnable startProvisioning = () -> experiment.setExperimentState(ExperimentState.PROVISIONING);

        if (!durationToStartTime.isNegative()) {


            scheduledExecutorService.schedule(startProvisioning, durationToStartTime.getSeconds() + 1, TimeUnit.SECONDS);
        } else {
            startProvisioning.run();
        }
    }

    @Override
    public CompletableFuture<Void> requestUpdate() {
        return submitJob(jobFactory.createUpdateExperimentJob(this.getExperiment()))
                .whenComplete((v, throwable) -> Platform.runLater(() -> {
                    if (throwable != null) {
                        LOG.warn("Update experiment failed: {}. Checking for experiment part states anyway.", throwable.getMessage(), throwable);
                    }
                    checkExperimentPartsStates();
                }));
    }

    @Override
    public void shareWithUsers(@Nonnull List<GeniUrn> selectedUsers, boolean registerSshKeys) {
        submitJob(jobFactory.createShareSliceJob(experiment, selectedUsers, registerSshKeys));
    }

    @Override
    public void unshareWithUsers(@Nonnull List<GeniUrn> selectedUsers, boolean unregisterSshKeys) {
        submitJob(jobFactory.createUnshareSliceJob(experiment, selectedUsers, unregisterSshKeys));
    }

    @Override
    public void updateSshKeys(@Nonnull List<UserSpec> userspec) {
        submitJob(jobFactory.createEditSshKeysJob(experiment, userspec));
    }

    @Override
    public void reloadOS(Sliver sliver) {
        submitJob(jobFactory.createReloadOSJob(experiment, sliver));
    }

    @Override
    public void createDiskImage(FXRspecNode node, String imageName, boolean global, boolean updatePrepare) {
        submitJob(jobFactory.createCreateDiskImageJob(experiment, node, imageName, global, updatePrepare));
    }

    @Override
    public void stop(Collection<ExperimentPart> experimentParts) {
        submitJob(jobFactory.createStopExperimentJob(experiment, experimentParts))
                .whenCompleteAsync((aVoid, throwable) -> {
                    if (throwable != null) {
                        LOG.warn("Stopping experiment failed: {}. Checking experiment part states anyway.",
                                throwable.getMessage(), throwable);
                    }
                    checkExperimentPartsStates();
                }, Platform::runLater);

        eventHandlerManager.dispatch(new ExperimentPartsEvent(this, ExperimentPartsEvent.STOPPED, experimentParts));
    }

    @Override
    public void stop() {
        stop(experiment.getPartsListCopy());
    }

    @Override
    public CompletableFuture<Boolean> testLinks() {
        assert experimentLinkTesterFactory != null;
        return submitJob(
                experimentLinkTesterFactory.createTestLinkJob(experiment)).whenComplete((success, throwable) -> {
            if (throwable == null) {
                LOG.debug("Link Test Job got result: " + success);
            } else {
                LOG.warn("Link Test Job Failed.", throwable.getMessage(), throwable);
            }
        });

    }

    @Override
    public void reboot(Sliver sliver) {
        submitJob(jobFactory.createRebootJob(experiment, sliver));
    }

    @Override
    public void reboot(ExperimentPart experimentPart) {
        assert (experiment.getPartsListCopy().contains(experimentPart));
        submitJob(jobFactory.createRebootJob(experimentPart));
    }

    @Override
    public void openConsole(Sliver sliver) {
        submitJob(jobFactory.createOpenConsoleJob(experiment, sliver));
    }

    @Override
    public void shareLan(FXRspecLink link, String sharedLanName) {
        submitJob(jobFactory.createShareLanJob(experiment, link, sharedLanName));
    }

    @Override
    public void unshareLan(FXRspecLink link, String sharedLanName) {
        submitJob(jobFactory.createUnshareLanJob(experiment, link, sharedLanName));
    }

    @Override
    public CompletableFuture<Collection<UserSpec>> fetchSliceMemberSshKeys() {
        return submitJob(jobFactory.createFetchSliceMemberSshKeysJob(experiment));
    }

    @Override
    public CompletableFuture<ExperimentTaskStatus> renew(Instant newExpirationTime) {
        return submitJob(jobFactory.createRenewExperimentJob(experiment, newExpirationTime));
    }

    @Override
    public void renewParts(Instant newExpirationTime, Collection<ExperimentPart> parts) {
        submitJob(jobFactory.createRenewExperimentJob(experiment, newExpirationTime, parts));
    }

    @Override
    public void log(@Nonnull LogOutput.LogEntry logEntry) {
        //forward to experiment
        experiment.log(logEntry);
    }


    private class ExpirationDetector implements Runnable {

        private final NextExperimentExpirationBinding nextExperimentExpirationBinding;
        private ScheduledFuture<?> previousScheduledExpiration = null;

        public ExpirationDetector() {
            this.nextExperimentExpirationBinding = new NextExperimentExpirationBinding(experiment);
            this.nextExperimentExpirationBinding.addListener(observable -> scheduleForNextExpiration());
        }


        @Override
        public void run() {
            final NextExperimentExpiration nextExpiration = nextExperimentExpirationBinding.get();
            if (nextExpiration.hasPartExpired()) {
                LOG.trace("A part of the experiment {} has expired!", experiment.getName());
                //(a part of) the experiment seems to have expired!
                experiment.setExperimentState(ExperimentState.EXPIRING);
            } else {
                scheduleForNextExpiration();
            }
        }

        private void scheduleForNextExpiration() {
            if (previousScheduledExpiration != null)
                previousScheduledExpiration.cancel(false);

            final NextExperimentExpiration nextExpiration = nextExperimentExpirationBinding.get();

            if (nextExpiration.getFirstExpirationTime() != null) {
                Duration durationToFirstExpiration =
                        Duration.between(Instant.now(), nextExpiration.getFirstExpirationTime());

                assert !durationToFirstExpiration.isNegative();

                if (!durationToFirstExpiration.isNegative()) {
                    LOG.debug("Scheduling next expiration check for {} in {} seconds", experiment.getName(), durationToFirstExpiration.getSeconds() + 1);

                    previousScheduledExpiration =
                            scheduledExecutorService.schedule(this, durationToFirstExpiration.getSeconds() + 1, TimeUnit.SECONDS);
                } else {
                    LOG.warn("NOT Scheduling next expiration check for {}: experiment expired {} seconds ago", durationToFirstExpiration.getSeconds());
                }
            } else if (!nextExpiration.hasPartExpired()) {
                LOG.debug("Could not find an expiration time. Trying again in 10 seconds");
                previousScheduledExpiration = scheduledExecutorService.schedule(this, 10, TimeUnit.SECONDS);
            }
        }
    }

    /**
     * This will also wait until the resource is ready
     */
    private void provisionResources() {
        assert experiment.getNewRequestRspecSource() != null;
        assert experiment.getSliceOrNull() != null;

        if (experiment.getSliceOrNull() != null) {
            experiment.getSliceOrNull().setRequestRspec(experiment.getNewRequestRspecSource());
        }

        ProvisionExperimentJob job = jobFactory.createProvisionExperimentJob(experiment);
        JobReport jobReport = job.getJobReport();
        experiment.addJobReport(jobReport);
        submitJob(job)
                .whenCompleteAsync((success, throwable) -> {
                    if (throwable != null) {
                        LOG.error("ProvisionExperimentJob for " + experiment.getName() + " failed: ", throwable);
                        experiment.setExperimentState(ExperimentState.FAILING);
                    } else {
                        if (Objects.equals(success, Boolean.TRUE))
                            experiment.setExperimentState(ExperimentState.WAIT_FOR_READY);
                        else {
                            LOG.debug("ProvisionExperimentJob for {} succeeded but returned success={}", experiment.getName(), success);
                            experiment.setExperimentState(ExperimentState.FAILING);
                        }
                    }
                }, Platform::runLater);

    }

    @Override
    public void waitForReady() {
        LOG.debug("experimentController.waitForReady()");
        WaitForOpStatusExperimentJob job = jobFactory.createWaitForReadyExperimentJob(experiment);
        JobReport jobReport = job.getJobReport();
        experiment.addJobReport(jobReport);
        submitJob(job).whenCompleteAsync((success, throwable) -> {
            if (throwable != null) {
                LOG.error("WaitForOpStatusExperimentJob for " + experiment.getName() + " failed: ", throwable);
                experiment.setExperimentState(ExperimentState.FAILING);
            } else {
                if (Objects.equals(success, Boolean.TRUE))
                    experiment.setExperimentState(ExperimentState.TESTING_CONNECTIVITY);
                else {
                    if (Objects.equals(job.isTimeout(), Boolean.TRUE)) {
                        LOG.debug("WaitForOpStatusExperimentJob for {} succeeded but there was a timeout", experiment.getName(), success);
                        experiment.setExperimentState(ExperimentState.TIMEOUT_WAITING);
                    } else {
                        LOG.debug("WaitForOpStatusExperimentJob for {} succeeded but returned success={}", experiment.getName(), success);
                        experiment.setExperimentState(ExperimentState.FAILING);
                    }
                }
            }
        }, Platform::runLater);
    }

    @Override
    public void setupSoftware() {
        submitJob(jobFactory.createSetupSoftwareExperimentJob(experiment))
                .whenCompleteAsync((success, throwable) -> {
                    if (throwable != null) {
                        LOG.error("SetupSoftwareExperimentJob for " + experiment.getName() + " failed: ", throwable);
                        experiment.setExperimentState(ExperimentState.FAILING);
                    } else {
                        if (Objects.equals(success, Boolean.TRUE))
                            experiment.setExperimentState(ExperimentState.READY);
                        else {
                            LOG.debug("SetupSoftwareExperimentJob for {} succeeded but returned success={}", experiment.getName(), success);
                            experiment.setExperimentState(ExperimentState.FAILING);
                        }
                    }
                }, Platform::runLater);
    }

    @Override
    public CompletableFuture<Boolean> rerunESpec(@Nullable ESpecStep startPosition,
                                                 @Nullable ESpecStep stopPosition) {
        LOG.debug("do rerunESpec start=" + startPosition + " stop=" + stopPosition);

        return submitJob(jobFactory.createRerunEspecJob(experiment, startPosition, stopPosition)).whenComplete((success, throwable) -> {
            if (throwable != null) {
                LOG.error("Rerun ESpec SetupSoftwareExperimentJob for " + experiment.getName() + " failed: ", throwable);
            } else {
                if (Objects.equals(success, Boolean.TRUE)) {
                    LOG.debug("Rerun ESpec SetupSoftwareExperimentJob for {} succeeded succesfully", experiment.getName());
                } else {
                    LOG.debug("Rerun ESpec SetupSoftwareExperimentJob for {} succeeded but returned success={}", experiment.getName(), success);
                }
            }
        });
    }

    @Nullable
    private Server getScsAuthority() {
        //explicitly set in experiment overrides both jFedPreferences and userauth based values
        if (experiment.getScs() != null && experiment.getScs().getServer() != null) {
            return experiment.getScs().getServer();
        }

        Server scsServer = jFedPreferences.getScsAuthority(testbedInfoSource);
        if (scsServer == null) {
            Integer scsId = geniUserProvider.getLoggedInGeniUser().getUserAuthorityServer() != null ?
                    geniUserProvider.getLoggedInGeniUser().getUserAuthorityServer().getDefaultScsId() :
                    null;
            if (scsId != null) {
                Service scsService = testbedInfoSource.getServiceById(scsId);
                assert scsService != null : "Could not find scs " + scsId;
                if (scsService != null) {
                    scsServer = scsService.getServer();
                    assert scsServer != null;
//                  GeniUrn defaultScsUrn = scsServer.getDefaultComponentManagerAsGeniUrn();
                }
            }
        }
        return scsServer;
    }

    private <V> CompletableFuture<V> submitJob(Job<V> job) {

        listeners.forEach(listener -> listener.onJobSubmitted(job));

        job.stateProperty().addListener(observable ->
                listeners.forEach(listener -> listener.onJobStateChanged(job, job.getState()))
        );

        CompletableFuture<V> completableFuture = new CompletableFuture<>();

        jobExecutorService.submit(()-> {
            try {
                completableFuture.complete(job.call());
            } catch (Exception e) {
                completableFuture.completeExceptionally(e);
            }
        });

        return completableFuture
                .whenComplete((v, throwable) -> listeners.forEach(listener -> listener.onJobFinished(job)));
    }

    @Override
    @Nonnull
    public Experiment getExperiment() {
        return experiment;
    }

    @Override
    public void addListener(ExperimentControllerListener listener) {
        this.listeners.add(listener);
    }

    @Override
    public void removeListener(ExperimentControllerListener listener) {
        this.listeners.remove(listener);
    }

    @Override
    public ExperimentConnectivityTesterFactory getExperimentConnectivityTesterFactory() {
        return experimentConnectivityTesterFactory;
    }

    @Override
    public void setExperimentConnectivityTesterFactory(ExperimentConnectivityTesterFactory experimentConnectivityTesterFactory) {
        this.experimentConnectivityTesterFactory = experimentConnectivityTesterFactory;
    }

    @Override
    public void setExperimentLinkTesterFactory(ExperimentLinkTesterFactory experimentLinkTesterFactory) {
        this.experimentLinkTesterFactory = experimentLinkTesterFactory;
    }

    @Override
    public WaitForReadyTimeoutHandler getWaitForReadyTimeoutHandler() {
        return waitForReadyTimeoutHandler;
    }

    @Override
    public void setWaitForReadyTimeoutHandler(WaitForReadyTimeoutHandler waitForReadyTimeoutHandler) {
        this.waitForReadyTimeoutHandler = waitForReadyTimeoutHandler;
    }

    @Override
    public <T extends ExperimentEvent> void addEventHandler(
            ExperimentEventType<T> type, ExperimentEventHandler<T> eventHandler) {
        eventHandlerManager.addExperimentEventHandler(type, eventHandler);
    }
}
